@@ -19,12 +19,13 @@ def setup_propagator(v: Tensor,
                      pml_width: Union[int, List[int]] = 20,
                      pml_freq: Optional[float] = None,
                      max_vel: Optional[float] = None,
-                     survey_pad:
-                       Optional[Union[int,
-                                      List[Optional[int]]]] = None,
+                     survey_pad: Optional[Union[int,
+                                          List[Optional[int]]]] = None,
                      origin: Optional[List[int]] = None,
                      nt: Optional[int] = None,
-                     model_gradient_sampling_interval: int = 1) -> Tuple[
+                     model_gradient_sampling_interval: int = 1,
+                     freq_taper_frac: float = 0.0,
+                     time_pad_frac: float = 0.0) -> Tuple[
                         Tensor, List[Tensor], Tensor, List[Tensor],
                         Tensor, Tensor, Tensor, Tensor, Tensor, Tensor,
                         float, float, float, int, int, int, int, int,
@@ -59,8 +60,7 @@ def setup_propagator(v: Tensor,
         if max_vel < v.abs().max().item():
             warnings.warn("max_vel is less than the actual maximum velocity")
     check_points_per_wavelength(v, pml_freq, dy, dx, dt)
-    v = pad_model(v, pml_width)
-    v = pad_model(v, [fd_pad for _ in range(4)], mode='constant')
+    v = pad_model(v, pad)
     for i in range(len(other_models)):
         other_models[i] = pad_model(other_models[i], pad,
                                     mode=other_model_modes[i])
@@ -73,7 +73,9 @@ def setup_propagator(v: Tensor,
                                    source_locations[..., 1],
                                    None] ** 2 * dt**2
         )
-        source_amplitudes = upsample(source_amplitudes, step_ratio)
+        source_amplitudes = upsample(source_amplitudes, step_ratio,
+                                     freq_taper_frac=freq_taper_frac,
+                                     time_pad_frac=time_pad_frac)
     else:
         sources_i = None
     if receiver_locations is not None:
@@ -83,8 +85,12 @@ def setup_propagator(v: Tensor,
         receivers_i = None
     n_batch = get_n_batch([source_locations, receiver_locations] +
                           list(wavefields))
-    ay, ax, by, bx = \
-        setup_pml(pml_width, fd_pad, dt, v, max_vel, pml_freq)
+    ay, by = \
+        setup_pml(pml_width[:2], fd_pad, dt, dy, v.shape[0], max_vel,
+                  v.dtype, v.device, pml_freq)
+    ax, bx = \
+        setup_pml(pml_width[2:], fd_pad, dt, dx, v.shape[1], max_vel,
+                  v.dtype, v.device, pml_freq)
     nt *= step_ratio
 
     if source_amplitudes is not None:
@@ -115,13 +121,16 @@ def setup_propagator(v: Tensor,
 
 
 def downsample_and_movedim(receiver_amplitudes: Tensor,
-                           step_ratio: int) -> Tensor:
+                           step_ratio: int, freq_taper_frac: float = 0.0,
+                           time_pad_frac: float = 0.0) -> Tensor:
     if receiver_amplitudes.numel() > 0:
         if receiver_amplitudes.device == torch.device('cpu'):
             receiver_amplitudes = torch.movedim(receiver_amplitudes, 1, -1)
         else:
             receiver_amplitudes = torch.movedim(receiver_amplitudes, 0, -1)
-        receiver_amplitudes = downsample(receiver_amplitudes, step_ratio)
+        receiver_amplitudes = downsample(receiver_amplitudes, step_ratio,
+                                         freq_taper_frac=freq_taper_frac,
+                                         time_pad_frac=time_pad_frac)
     return receiver_amplitudes
 
 
@@ -252,9 +261,9 @@ def get_n_batch(args: List[Optional[Tensor]]) -> int:
 
 def pad_model(model: Tensor, pad: List[int],
               mode: str = "replicate") -> Tensor:
-    return torch.nn.functional.pad(model[None],
+    return torch.nn.functional.pad(model[None, None],
                                    (pad[2], pad[3], pad[0], pad[1]),
-                                   mode=mode)[0]
+                                   mode=mode)[0, 0]
 
 
 def pad_locations(locations: Tensor, pad: List[int]) -> Tensor:
@@ -267,23 +276,145 @@ def location_to_index(locations: Tensor, nx: int) -> Tensor:
     return locations[..., 0] * nx + locations[..., 1]
 
 
-def upsample(signal: Tensor, step_ratio: int) -> Tensor:
+def cosine_taper_end(signal: Tensor, n_taper: int) -> Tensor:
+    """Tapers the end of the final dimension of a Tensor using a cosine.
+
+    A half period, shifted and scaled to taper from 1 to 0, is used.
+
+    Args:
+        signal:
+            The Tensor that will have its final dimension tapered.
+        n_taper:
+            The length of the cosine taper, in number of samples.
+
+    Returns:
+        The tapered signal.
+    """
+    taper = torch.ones(signal.shape[-1], dtype=signal.real.dtype,
+                       device=signal.device)
+    taper[len(taper)-n_taper:] = (
+        torch.cos(torch.arange(n_taper, dtype=signal.real.dtype,
+                               device=signal.device) /
+                  (n_taper - 1) * torch.pi) + 1
+    ) / 2
+    return signal * taper
+
+
+def zero_last_element_of_final_dimension(signal: Tensor) -> Tensor:
+    zeroer = torch.ones(signal.shape[-1], dtype=signal.real.dtype,
+                        device=signal.device)
+    zeroer[-1] = 0
+    return signal * zeroer
+
+
+def upsample(signal: Tensor, step_ratio: int, freq_taper_frac: float = 0.0,
+             time_pad_frac: float = 0.0) -> Tensor:
+    """Upsamples the final dimension of a Tensor by a factor.
+
+    Low-pass upsampling is used to produce an upsampled signal without
+    introducing higher frequencies than were present in the input. The
+    Nyquist frequency of the input will be zeroed.
+
+    Args:
+        signal:
+            The Tensor that will have its final dimension upsampled.
+        step_ratio:
+            The integer factor by which the signal will be upsampled.
+            The input signal is returned if this is 1 (freq_taper_frac
+            and time_pad_frac will be ignored).
+        freq_taper_frac:
+            A float specifying the fraction of the end of the signal
+            amplitude in the frequency domain to cosine taper. This
+            might be useful to reduce ringing. A value of 0.1 means
+            that the top 10% of frequencies will be tapered before
+            upsampling. Default 0.0 (no tapering).
+        time_pad_frac:
+            A float specifying the amount of zero padding that will
+            be added to the signal before upsampling and removed
+            afterwards, as a fraction of the length of the final
+            dimension of the input signal. This might be useful to reduce
+            wraparound artifacts. A value of 0.1 means that zero padding
+            of 10% of the length of the signal will be used. Default 0.0.
+
+    Returns:
+        The signal after upsampling.
+    """
     if step_ratio == 1:
         return signal
+    if time_pad_frac > 0.0:
+        n_time_pad = int(time_pad_frac * signal.shape[-1])
+        signal = torch.nn.functional.pad(signal, (0, n_time_pad))
+    else:
+        n_time_pad = 0
     nt = signal.shape[-1]
     up_nt = nt * step_ratio
-    return torch.fft.irfft(torch.nn.functional.pad(torch.fft.rfft(signal),
-                                                   (0, up_nt//2+1 - nt//2+1)),
-                           n=up_nt)
+    signal_f = torch.fft.rfft(signal)
+    if freq_taper_frac > 0.0:
+        freq_taper_len = int(freq_taper_frac * signal_f.shape[-1])
+        signal_f = cosine_taper_end(signal_f, freq_taper_len)
+    else:
+        # Set Nyquist frequency to zero
+        signal_f = zero_last_element_of_final_dimension(signal_f)
+    signal_f = torch.nn.functional.pad(signal_f,
+                                       (0, up_nt//2+1 - nt//2+1))
+    signal = torch.fft.irfft(signal_f, n=up_nt)
+    if time_pad_frac > 0.0:
+        signal = signal[..., :signal.shape[-1] - n_time_pad * step_ratio]
+    return signal
 
 
-def downsample(signal: Tensor, step_ratio: int) -> Tensor:
+def downsample(signal: Tensor, step_ratio: int, freq_taper_frac: float = 0.0,
+               time_pad_frac: float = 0.0) -> Tensor:
+    """Downsamples the final dimension of a Tensor by a factor.
+
+    Frequencies higher than or equal to the Nyquist frequency of the
+    downsampled signal will be zeroed before downsampling.
+
+    Args:
+        signal:
+            The Tensor that will have its final dimension downsampled.
+        step_ratio:
+            The integer factor by which the signal will be downsampled.
+            The input signal is returned if this is 1 (freq_taper_frac
+            and time_pad_frac will be ignored).
+        freq_taper_frac:
+            A float specifying the fraction of the end of the signal
+            amplitude in the frequency domain to cosine taper. This
+            might be useful to reduce ringing. A value of 0.1 means
+            that the top 10% of frequencies will be tapered after
+            downsampling. Default 0.0 (no tapering).
+        time_pad_frac:
+            A float specifying the amount of zero padding that will
+            be added to the signal before downsampling and removed
+            afterwards, as a fraction of the length of the final
+            dimension of the output signal. This might be useful to reduce
+            wraparound artifacts. A value of 0.1 means that zero padding
+            of 10% of the length of the output signal will be used.
+            Default 0.0.
+
+    Returns:
+        The signal after downsampling.
+    """
     if step_ratio == 1:
         return signal
+    if time_pad_frac > 0.0:
+        n_time_pad = int(time_pad_frac * (signal.shape[-1] // step_ratio))
+        signal = torch.nn.functional.pad(signal, (0, n_time_pad * step_ratio))
+    else:
+        n_time_pad = 0
     nt = signal.shape[-1]
     down_nt = nt // step_ratio
-    return torch.fft.irfft(torch.fft.rfft(signal)[..., :down_nt//2+1],
-                           n=down_nt)
+    signal_f = torch.fft.rfft(signal)[..., :down_nt//2+1]
+    if freq_taper_frac > 0.0:
+        freq_taper_len = int(freq_taper_frac * signal_f.shape[-1])
+        signal_f = cosine_taper_end(signal_f, freq_taper_len)
+    else:
+        # Set Nyquist frequency to zero
+        signal_f = zero_last_element_of_final_dimension(signal_f)
+    signal = torch.fft.irfft(signal_f, n=down_nt)
+    if time_pad_frac > 0.0:
+        signal = signal[..., :signal.shape[-1] - n_time_pad]
+    return signal
 
 
 def convert_to_contiguous(tensor: Optional[Tensor]) -> Tensor:
@@ -299,7 +430,7 @@ def extract_survey(models: List[Tensor],
                    wavefields: List[Optional[Tensor]],
                    origin: Optional[List[int]],
                    pad: List[int]) -> Tuple[List[Tensor],
-                                                List[Optional[Tensor]]]:
+                                            List[Optional[Tensor]]]:
     check_locations_are_within_model(models[0].shape, locations)
     if any([wavefield is not None for wavefield in wavefields]):
         survey_extents = get_survey_extents_from_wavefields(
@@ -563,6 +694,30 @@ def extract_locations(locations: List[Optional[Tensor]],
 
 def cfl_condition(dy: float, dx: float, dt: float, max_vel: float,
                   eps: float = 1e-15) -> Tuple[float, int]:
+    """Calculates the time step interval to obey the CFL condition.
+
+    The output time step will be a factor of the input time step.
+
+    Args:
+        dy:
+            The grid spacing in the first dimension.
+        dx:
+            The grid spacing in the second dimension.
+        dt:
+            The time step interval.
+        max_vel:
+            The maximum wave speed in the model.
+        eps:
+            A small quantity to prevent division by zero. Default 1e-15.
+
+    Returns:
+        Tuple[float, int]:
+
+            inner_dt:
+                A time step interval that obeys the CFL condition.
+            step_ratio:
+                The integer dt / inner_dt.
+    """
     max_dt = (0.6 / math.sqrt(1 / dy**2 + 1 / dx**2) /
               (max_vel**2 + eps)) * max_vel
     step_ratio = int(math.ceil((abs(dt) / (max_dt**2 + eps)) * max_dt))
@@ -570,55 +725,32 @@ def cfl_condition(dy: float, dx: float, dt: float, max_vel: float,
     return inner_dt, step_ratio
 
 
-def setup_pml(pml_width: List[int], fd_pad: int, dt: float, v: Tensor,
-              max_vel: float, pml_freq: Optional[float],
-              eps: float = 1e-9) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
+def setup_pml(pml_width: List[int], fd_pad: int, dt: float, dx: float, n: int,
+              max_vel: float, dtype: torch.dtype, device: torch.device,
+              pml_freq: Optional[float], 
+              eps: float = 1e-9) -> Tuple[Tensor, Tensor]:
     R = 0.001
     if pml_freq is None:
         pml_freq = 25.0
-    device = v.device
-    dtype = v.dtype
-    a = []
-    b = []
-    for width in pml_width:
+    alpha0 = math.pi * pml_freq
+    a = math.exp(-alpha0 * abs(dt)) * torch.ones(n, device=device, dtype=dtype)
+    b = torch.zeros(n, device=device, dtype=dtype)
+    for side, width in enumerate(pml_width):
         if width == 0:
-            a.append(torch.zeros(0, device=device, dtype=dtype))
-            b.append(torch.zeros(0, device=device, dtype=dtype))
             continue
-        sigma = (
-            3
-            * max_vel
-            * math.log((1 + R) / 2 * width)
-            * (torch.arange(width, device=device, dtype=dtype)
-               / width) ** 2
-        )
-        alpha = (
-            2 * math.pi * pml_freq
-            * (1 - torch.arange(width, device=device, dtype=dtype)
-               / width)
+        pml_frac = (
+            torch.arange(width + fd_pad + 1, device=device, dtype=dtype)
+            / width
         )
+        sigma = -3 * max_vel * math.log(R) / (2 * width * dx) * pml_frac**2
+        alpha = alpha0 * (1 - pml_frac)
         sigmaalpha = sigma + alpha
-        a.append(torch.exp(-sigmaalpha * abs(dt)))
-        b.append(sigma / (sigmaalpha**2 + eps) * sigmaalpha * (a[-1] - 1))
-    a0 = math.exp(-2 * math.pi * pml_freq * abs(dt))
-    ay = a0 * torch.ones(v.shape[0], device=device, dtype=dtype)
-    ax = a0 * torch.ones(v.shape[1], device=device, dtype=dtype)
-    by = torch.zeros(v.shape[0], device=device, dtype=dtype)
-    bx = torch.zeros(v.shape[1], device=device, dtype=dtype)
-    ay[fd_pad:fd_pad + pml_width[0]] = a[0].flip(0)
-    ay[:fd_pad] = 0
-    ay[len(ay) - pml_width[1] - fd_pad:-fd_pad] = a[1]
-    ay[-fd_pad:] = 0
-    ax[fd_pad:fd_pad + pml_width[2]] = a[2].flip(0)
-    ax[:fd_pad] = 0
-    ax[len(ax) - pml_width[3] - fd_pad:-fd_pad] = a[3]
-    ax[-fd_pad:] = 0
-    by[fd_pad:fd_pad + pml_width[0]] = b[0].flip(0)
-    by[:fd_pad] = 0
-    by[len(by) - pml_width[1] - fd_pad:-fd_pad] = b[1]
-    by[-fd_pad:] = 0
-    bx[fd_pad:fd_pad + pml_width[2]] = b[2].flip(0)
-    bx[:fd_pad] = 0
-    bx[len(bx) - pml_width[3] - fd_pad:-fd_pad] = b[3]
-    bx[-fd_pad:] = 0
-    return ay, ax, by, bx
+        a_side = torch.exp(-sigmaalpha * abs(dt))
+        b_side = sigma / (sigmaalpha**2 + eps) * sigmaalpha * (a_side - 1)
+        if side == 0:
+            a[:fd_pad + pml_width[0] + 1] = a_side.flip(0)
+            b[:fd_pad + pml_width[0] + 1] = b_side.flip(0)
+        else:
+            a[-(fd_pad + pml_width[1] + 1):] = a_side
+            b[-(fd_pad + pml_width[1] + 1):] = b_side
+    return a, b
