@@ -2,10 +2,11 @@
 import math
 from typing import Optional
 import torch
+from torch import Tensor
 
 
 def ricker(freq: float, length: int, dt: float, peak_time: float,
-           dtype: Optional[torch.dtype] = None):
+           dtype: Optional[torch.dtype] = None) -> Tensor:
     """Return a Ricker wavelet with the specified central frequency.
 
     Args:
